import os

from lazy_loader import attach as _attach

from napari._check_numpy_version import limit_numpy1x_threads_on_macos_arm

try:
    from napari._version import version as __version__
except ImportError:
    __version__ = 'not-installed'

# Allows us to use pydata/sparse arrays as layer data
os.environ.setdefault('SPARSE_AUTO_DENSIFY', '1')
limit_numpy1x_threads_on_macos_arm()


def _check_installation_path():  # pragma: no cover
    """Check for installation path conflicts.

    Check if napari is present in site-packages. If napari is installed in editable mode,
    notify the user of a the conflict that napari is also in site-packages.
    """
    import sys
    from pathlib import Path

    if 'pytest' in sys.modules:
        # pytest is running, skip the check
        return

    napari_installation_path = Path(__file__).absolute().parent.parent
    if napari_installation_path.name == 'site-packages':
        # napari is installed in non-editable mode
        return

    import numpy as np

    # Use numpy location to determine a site-packages path
    site_packages_path = Path(np.__file__).absolute().parent.parent
    if site_packages_path.name != 'site-packages':
        # numpy is not installed in site-packages
        return

    napari_site_packages_path = site_packages_path / 'napari'
    napari_builtins_package_path = site_packages_path / 'napari_builtins'

    path_text = ''
    if napari_site_packages_path.exists():
        path_text += (
            f'Path to a napari directory: {napari_site_packages_path}.\n'
        )
    if napari_builtins_package_path.exists():
        path_text += f'Path to a napari_builtins directory: {napari_builtins_package_path}.\n'

    if path_text:
        text = (
            'Mix of local and non local installation detected.\n'
            'Napari is installed in editable mode but also found napari '
            'directory in site-packages.\n'
            f'{path_text}'
            'Mix of local and non local installation is leading '
            'to hard to understand errors. '
            'See https://napari.org/stable/troubleshooting.html#mixed-napari-installations for more details.'
        )
        raise RuntimeError(text)


_check_installation_path()

del limit_numpy1x_threads_on_macos_arm
del os

# Add everything that needs to be accessible from the napari namespace here.
_proto_all_ = [
    '__version__',
    'components',
    'experimental',
    'layers',
    'qt',
    'types',
    'viewer',
    'utils',
]

_submod_attrs = {
    '_event_loop': ['run'],
    'plugins.io': ['save_layers'],
    'utils': ['sys_info'],
    'utils.notifications': ['notification_manager'],
    'view_layers': [
        'view_image',
        'view_labels',
        'view_path',
        'view_points',
        'view_shapes',
        'view_surface',
        'view_tracks',
        'view_vectors',
        'imshow',
    ],
    'viewer': ['Viewer', 'current_viewer'],
}

# All imports in __init__ are hidden inside of `__getattr__` to prevent
# importing the full chain of packages required when calling `import napari`.
#
# This has the biggest implications for running `napari` on the command line
# (or running `python -m napari`) since `napari.__init__` gets imported
# on the way to `napari.__main__`. Importing everything here has the
# potential to take a second or more, so we definitely don't want to import it
# just to access the CLI (which may not actually need any of the imports)


__getattr__, __dir__, __all__ = _attach(
    __name__, submodules=_proto_all_, submod_attrs=_submod_attrs
)
del _attach



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QSpinBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Labels
from napari.utils.translations import trans


class QtNdimSpinBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer number of
    editable dimensions attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    ndim_spinbox : qtpy.QtWidgets.QSpinBox
        Spinbox to control the number of editable dimensions of label layer.
    ndim_spinbox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the number of editable dimensions chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.n_edit_dimensions.connect(
            self._on_n_edit_dimensions_change
        )
        self._layer.events.data.connect(self._on_data_change)

        # Setup widgets
        ndim_sb = QSpinBox()
        self.ndim_spinbox = ndim_sb
        ndim_sb.setToolTip(trans._('Number of dimensions for label editing'))
        ndim_sb.valueChanged.connect(self.change_n_edit_dim)
        ndim_sb.setMinimum(2)
        ndim_sb.setMaximum(self._layer.ndim)
        ndim_sb.setSingleStep(1)
        ndim_sb.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._on_n_edit_dimensions_change()

        self.ndim_spinbox_label = QtWrappedLabel(trans._('n edit dim:'))

    def change_n_edit_dim(self, value: int) -> None:
        """Change the number of editable dimensions of label layer.
        Parameters
        ----------
        value : int
            The number of editable dimensions to set.
        """
        self._layer.n_edit_dimensions = value
        self.ndim_spinbox.clearFocus()

        # TODO: Check how to decouple this
        self.parent().setFocus()

    def _on_n_edit_dimensions_change(self) -> None:
        """Receive layer model n-dim mode change event and update the checkbox."""
        with qt_signals_blocked(self.ndim_spinbox):
            value = self._layer.n_edit_dimensions
            self.ndim_spinbox.setValue(int(value))
            # TODO: Check how to decouple this
            self.parent()._set_polygon_tool_state()

    def _on_data_change(self) -> None:
        self.ndim_spinbox.setMaximum(self._layer.ndim)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.ndim_spinbox_label, self.ndim_spinbox)]



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QSlider,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Tracks
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtHeadLengthSliderControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the current head length
    attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Tracks
        An instance of a napari Tracks layer.

    Attributes
    ----------
    head_length_slider : qtpy.QtWidgets.QSlider
        Slider controlling head length of the layer.
    head_length_slider_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the head length chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Tracks) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.head_length.connect(self._on_head_length_change)

        # Setup widgets
        # slider for track head length
        self.head_length_slider = QSlider(Qt.Orientation.Horizontal)
        self.head_length_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.head_length_slider.setMinimum(0)
        self.head_length_slider.setMaximum(self._layer._max_length)
        self.head_length_slider.setSingleStep(1)
        connect_setattr(
            self.head_length_slider.valueChanged, self._layer, 'head_length'
        )

        self.head_length_slider_label = QtWrappedLabel(trans._('head length:'))

    def _on_head_length_change(self) -> None:
        """Receive layer model track line width change event and update slider."""
        with qt_signals_blocked(self.head_length_slider):
            value = self._layer.head_length
            if value > self.head_length_slider.maximum():
                self.head_length_slider.setMaximum(self._layer._max_length)
            self.head_length_slider.setValue(value)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.head_length_slider_label, self.head_length_slider)]



import numpy as np
import numpy.testing as npt
import pytest

from napari.layers.shapes._shape_list import ShapeList
from napari.layers.shapes._shapes_models import Path, Polygon, Rectangle


@pytest.fixture
def shape_li():
    """Fixture for ShapeList."""
    shape1 = Rectangle(np.array([[0, 0], [10, 10]]))
    shape2 = Rectangle(np.array([[10, 0], [0, 10]]))
    shape3 = Rectangle(np.array([[20, 20], [30, 30]]))
    shape4 = Rectangle(np.array([[-10, -10], [0, 0]]))
    res = ShapeList()
    res.add([shape1, shape2, shape3, shape4])
    return res


@pytest.fixture
def shape_li_3d():
    """Fixture for ShapeList."""
    shape1 = Rectangle(
        np.array([[0, 0, 0], [0, 10, 0], [0, 10, 10], [0, 0, 10]])
    )
    shape2 = Rectangle(
        np.array([[1, 10, 0], [1, 10, 10], [1, 0, 10], [1, 0, 0]])
    )
    shape3 = Rectangle(
        np.array([[0, 20, 20], [0, 20, 30], [0, 30, 30], [0, 30, 20]])
    )
    shape4 = Rectangle(
        np.array([[1, -10, -10], [1, -10, 0], [1, 0, 0], [1, 0, -10]])
    )
    res = ShapeList()
    res.add([shape1, shape2, shape3, shape4])
    return res


def test_empty_shape_list():
    """Test instantiating empty ShapeList."""
    shape_list = ShapeList()
    assert len(shape_list.shapes) == 0


def test_adding_to_shape_list():
    """Test adding shapes to ShapeList."""
    np.random.seed(0)
    data = 20 * np.random.random((4, 2))
    shape = Rectangle(data)
    shape_list = ShapeList()

    shape_list.add(shape)
    assert len(shape_list.shapes) == 1
    assert shape_list.shapes[0] == shape


def test_reset_bounding_box_rotation():
    """Test if rotating shape resets bounding box."""
    shape = Rectangle(np.array([[0, 0], [10, 10]]))
    shape_list = ShapeList()
    shape_list.add(shape)
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, np.array([[[-0.5, -0.5]], [[10.5, 10.5]]])
    )
    shape_list.rotate(0, 45, (5, 5))
    p = 5 * np.sqrt(2) + 0.5
    npt.assert_array_almost_equal(
        shape.bounding_box, np.array([[5 - p, 5 - p], [5 + p, 5 + p]])
    )
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, shape.bounding_box[:, np.newaxis, :]
    )


def test_reset_bounding_box_shift():
    """Test if shifting shape resets bounding box."""
    shape = Rectangle(np.array([[0, 0], [10, 10]]))
    shape_list = ShapeList()
    shape_list.add(shape)
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, shape.bounding_box[:, np.newaxis, :]
    )
    shape_list.shift(0, np.array([5, 5]))
    npt.assert_array_almost_equal(
        shape.bounding_box, np.array([[4.5, 4.5], [15.5, 15.5]])
    )
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, shape.bounding_box[:, np.newaxis, :]
    )


def test_reset_bounding_box_scale():
    """Test if scaling shape resets the bounding box."""
    shape = Rectangle(np.array([[0, 0], [10, 10]]))
    shape_list = ShapeList()
    shape_list.add(shape)
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, shape.bounding_box[:, np.newaxis, :]
    )
    shape_list.scale(0, 2, (5, 5))
    npt.assert_array_almost_equal(
        shape.bounding_box, np.array([[-5.5, -5.5], [15.5, 15.5]])
    )
    npt.assert_array_almost_equal(
        shape_list._bounding_boxes, shape.bounding_box[:, np.newaxis, :]
    )


def test_shape_list_outline():
    """Test ShapeList outline method."""
    np.random.seed(0)
    data = 20 * np.random.random((4, 2))
    shape = Rectangle(data)
    shape_list = ShapeList()

    shape_list.add(shape)

    # Check passing an int
    outline_by_index = shape_list.outline(0)
    assert isinstance(outline_by_index, tuple)

    # Check passing a list
    outline_by_index_list = shape_list.outline([0])
    assert isinstance(outline_by_index_list, tuple)

    # Check return value for `int` and `list` are the same
    for value_by_idx, value_by_idx_list in zip(
        outline_by_index, outline_by_index_list, strict=False
    ):
        assert np.array_equal(value_by_idx, value_by_idx_list)

    # Check passing a `numpy.int_` (`numpy.int32/64` depending on platform)
    outline_by_index_np = shape_list.outline(np.int_(0))
    assert isinstance(outline_by_index_np, tuple)

    # Check return value for `int` and `numpy.int_` are the same
    for value_by_idx, value_by_idx_np in zip(
        outline_by_index, outline_by_index_np, strict=False
    ):
        assert np.array_equal(value_by_idx, value_by_idx_np)


def test_shape_list_outline_two_shapes():
    shape1 = Polygon([[0, 0], [0, 10], [10, 10], [10, 0]])
    shape2 = Polygon([[20, 20], [20, 30], [30, 30], [30, 20]])
    shape_list = ShapeList()
    shape_list.add([shape1, shape2])

    # check if the outline contains triangle with vertex of number 16

    triangles = shape_list.outline([0, 1])[2]
    assert np.any(triangles == 16)


def test_nD_shapes():
    """Test adding shapes to ShapeList."""
    np.random.seed(0)
    shape_list = ShapeList()
    data = 20 * np.random.random((6, 3))
    data[:, 0] = 0
    shape_a = Polygon(data)
    shape_list.add(shape_a)

    data = 20 * np.random.random((6, 3))
    data[:, 0] = 1
    shape_b = Path(data)
    shape_list.add(shape_b)

    assert len(shape_list.shapes) == 2
    assert shape_list.shapes[0] == shape_a
    assert shape_list.shapes[1] == shape_b

    assert shape_list._vertices.shape[1] == 2
    assert shape_list._mesh.vertices.shape[1] == 2

    shape_list.ndisplay = 3
    assert shape_list._vertices.shape[1] == 3
    assert shape_list._mesh.vertices.shape[1] == 3


@pytest.mark.parametrize('attribute', ['edge', 'face'])
def test_bad_color_array(attribute):
    """Test adding shapes to ShapeList."""
    np.random.seed(0)
    data = 20 * np.random.random((4, 2))
    shape = Rectangle(data)
    shape_list = ShapeList()

    shape_list.add(shape)

    # test setting color with a color array of the wrong shape
    bad_color_array = np.array([[0, 0, 0, 1], [1, 1, 1, 1]])
    with pytest.raises(ValueError, match='must have shape'):
        setattr(shape_list, f'{attribute}_color', bad_color_array)


def test_inside():
    shape1 = Polygon(np.array([[0, 0, 0], [0, 1, 0], [0, 1, 1], [0, 0, 1]]))
    shape2 = Polygon(np.array([[1, 0, 0], [1, 1, 0], [1, 1, 1], [1, 0, 1]]))
    shape3 = Polygon(np.array([[2, 0, 0], [2, 1, 0], [2, 1, 1], [2, 0, 1]]))

    shape_list = ShapeList()
    shape_list.add([shape1, shape2, shape3])
    shape_list.slice_key = (1,)
    assert shape_list.inside((0.5, 0.5)) == 1


def test_visible_shapes_4d():
    """Test _visible_shapes with 4D data like those from OME-Zarr/OMERO."""
    shape1 = Polygon(
        np.array(
            [
                [0, 0, 10, 10],
                [0, 0, 10, 50],
                [0, 0, 50, 50],
                [0, 0, 50, 10],
            ]
        )
    )
    shape2 = Polygon(
        np.array(
            [
                [0, 1, 10, 10],
                [0, 1, 10, 50],
                [0, 1, 50, 50],
                [0, 1, 50, 10],
            ]
        )
    )
    shape3 = Polygon(
        np.array(
            [
                [0, 0, 10, 10],
                [0, 0, 10, 50],
                [0, 1, 50, 50],
                [0, 1, 50, 10],
            ]
        )
    )

    shape_list = ShapeList()
    # set slice_key first to avoid empty array broadcasting error
    shape_list.slice_key = np.array([0, 0])
    shape_list.add([shape1, shape2, shape3])

    # at (0,0) - should show shape1 and shape3
    shape_list.slice_key = np.array([0, 0])
    visible = shape_list._visible_shapes
    assert len(visible) == 2
    visible_shapes = [v[1] for v in visible]
    assert shape1 in visible_shapes
    assert shape3 in visible_shapes

    # at (0,1) - should show shape2 and shape3
    shape_list.slice_key = np.array([0, 1])
    visible = shape_list._visible_shapes
    assert len(visible) == 2
    visible_shapes = [v[1] for v in visible]
    assert shape2 in visible_shapes
    assert shape3 in visible_shapes


def test_remove_shape(shape_li):
    """Test removing shapes from ShapeList."""
    initial_length = len(shape_li.shapes)
    assert initial_length == 4
    assert shape_li._vertices_index.shape[0] == 5

    # Remove a shape by index
    shape_li.remove(1)
    assert len(shape_li.shapes) == initial_length - 1
    assert shape_li._vertices_index.shape[0] == 4


def test_edit_shape_simple(shape_li):
    """Test editing shapes in ShapeList."""
    initial_shape = shape_li.shapes[0]
    assert isinstance(initial_shape, Rectangle)

    # Edit the first shape
    shape_li.edit(1, np.array([[5, 5], [15, 15]]))

    # Check if the shape has been updated
    assert len(shape_li.shapes) == 4
    npt.assert_array_equal(
        shape_li.shapes[1].data,
        np.array([[5.0, 5.0], [15.0, 5.0], [15.0, 15.0], [5.0, 15.0]]),
    )


def test_edit_shape_triangle(shape_li):
    """Test editing a shape to a triangle in ShapeList."""
    initial_shape = shape_li.shapes[0]
    assert isinstance(initial_shape, Rectangle)

    # Edit the first shape to a triangle
    shape_li.edit(1, np.array([[5, 5], [15, 5], [10, 15]]), new_type=Polygon)

    # Check if the shape has been updated
    assert len(shape_li.shapes) == 4
    npt.assert_array_equal(
        shape_li.shapes[1].data,
        np.array([[5.0, 5.0], [15.0, 5.0], [10.0, 15.0]]),
    )

    assert shape_li._mesh.triangles.shape[0] == 40


def test_edit_shape_pentagon(shape_li):
    """Test editing a shape to a polygon in ShapeList."""
    initial_shape = shape_li.shapes[0]
    assert isinstance(initial_shape, Rectangle)

    # Edit the first shape to a polygon
    shape_li.edit(
        1,
        np.array([[0, 0], [10, 0], [10, 10], [8, 15], [0, 10]]),
        new_type=Polygon,
    )

    # Check if the shape has been updated
    assert len(shape_li.shapes) == 4
    npt.assert_array_equal(
        shape_li.shapes[1].data,
        np.array([[0, 0], [10, 0], [10, 10], [8, 15], [0, 10]]),
    )
    assert shape_li._mesh.triangles.shape[0] == 43


@pytest.mark.parametrize(
    'new_color',
    [
        np.array([1, 0, 0, 1]),
        np.array([[1, 0, 0, 1]]),
        np.array([[1, 0, 0, 1]] * 4),
    ],
)
def test_update_face_color(shape_li, new_color):
    """Test updating face color of shapes in ShapeList."""
    # Initial face color
    initial_edge_color = shape_li.edge_color
    assert initial_edge_color.shape == (4, 4)

    # Update face color
    expected_color = np.array([[1, 0, 0, 1]] * 4)  # Red color
    shape_li.update_face_colors(range(4), new_color)

    # Check if the face color has been updated
    npt.assert_array_equal(shape_li.face_color, expected_color)
    assert (
        np.count_nonzero(
            np.all(
                shape_li._mesh.displayed_triangles_colors == expected_color[0],
                axis=1,
            )
        )
        == 8
    )

    # Check if the edge color remains unchanged
    assert np.all(shape_li.edge_color == initial_edge_color)


@pytest.mark.parametrize(
    'new_color',
    [
        np.array([1, 0, 0, 1]),
        np.array([[1, 0, 0, 1]]),
        np.array([[1, 0, 0, 1]] * 4),
    ],
)
def test_update_edge_color(shape_li, new_color):
    """Test updating edge color of shapes in ShapeList."""
    # Initial edge color
    initial_face_color = shape_li.face_color
    assert initial_face_color.shape == (4, 4)

    # Update edge color
    expected_color = np.array([[1, 0, 0, 1]] * 4)
    shape_li.update_edge_colors(range(4), new_color)
    # Check if the edge color has been updated
    npt.assert_array_equal(shape_li.edge_color, expected_color)
    assert (
        np.count_nonzero(
            np.all(
                shape_li._mesh.displayed_triangles_colors == expected_color[0],
                axis=1,
            )
        )
        == 32
    )

    # Check if the face color remains unchanged
    assert np.all(shape_li.face_color == initial_face_color)


def test_multi_layer_data(shape_li_3d):
    shape_li_3d.slice_key = (1,)
    assert shape_li_3d._mesh.displayed_triangles_colors.shape[0] == 20


LAYER_COUNT = 16


@pytest.fixture
def multi_z_rectangles() -> list[Rectangle]:
    width = 6
    edge = width - 2
    return [  # create 256 boxes in a grid
        Rectangle(
            np.array(
                [
                    [z, y * width, x * width],
                    [z, y * width + edge, x * width],
                    [z, y * width + edge, x * width + edge],
                    [z, y * width, x * width + edge],
                ]
            )
        )
        for x in range(4)
        for y in range(4)
        for z in range(LAYER_COUNT)
    ]


@pytest.fixture
def simple_rectangle():
    return Rectangle([(0, 0), (6, 0), (6, 6), (0, 6)])


@pytest.fixture
def triangles_slice(simple_rectangle):
    triang = np.concatenate(
        [
            simple_rectangle._face_triangles,
            simple_rectangle._edge_triangles
            + simple_rectangle.face_vertices_count,
        ]
    )

    return np.concatenate(
        [
            triang + i * simple_rectangle.vertices_count
            for i in range(0, 16 * LAYER_COUNT, LAYER_COUNT)
        ]
    )


@pytest.mark.parametrize('slice_', list(range(LAYER_COUNT)))
def test_proper_shape_position(
    multi_z_rectangles, slice_, triangles_slice, simple_rectangle
):
    sl = ShapeList(multi_z_rectangles)
    sl.ndisplay = 2
    sl.slice_key = (slice_,)
    assert sl._mesh.displayed_triangles_colors.shape[0] == 160
    npt.assert_array_equal(
        sl._mesh.displayed_triangles,
        triangles_slice + slice_ * simple_rectangle.vertices_count,
    )



from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay
from napari._vispy.visuals.text import Text
from napari.components._viewer_constants import CanvasPosition


class VispyTextOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    """Text overlay."""

    def __init__(self, *, viewer, overlay, parent=None) -> None:
        super().__init__(
            node=Text(pos=(0, 0)),
            viewer=viewer,
            overlay=overlay,
            parent=parent,
        )

        self.node.anchors = ('left', 'bottom')

        self.overlay.events.text.connect(self._on_text_change)
        self.overlay.events.color.connect(self._on_color_change)
        self.overlay.events.font_size.connect(self._on_position_change)

        self.reset()

    def _on_text_change(self):
        self.node.text = self.overlay.text
        self._on_position_change()

    def _on_visible_change(self):
        # ensure that dpi is updated when the scale bar is visible
        # this does not need to run _on_position_change because visibility
        # is already connected to the canvas callback by the canvas itself
        self._on_text_change()
        return super()._on_visible_change()

    def _on_color_change(self):
        self.node.color = self.overlay.color

    def _on_position_change(self, event=None):
        position = self.overlay.position
        anchors = ('left', 'bottom')
        if position == CanvasPosition.TOP_LEFT:
            anchors = ('left', 'bottom')
        elif position == CanvasPosition.TOP_RIGHT:
            anchors = ('right', 'bottom')
        elif position == CanvasPosition.TOP_CENTER:
            anchors = ('center', 'bottom')
        elif position == CanvasPosition.BOTTOM_RIGHT:
            anchors = ('right', 'top')
        elif position == CanvasPosition.BOTTOM_LEFT:
            anchors = ('left', 'top')
        elif position == CanvasPosition.BOTTOM_CENTER:
            anchors = ('center', 'top')

        self.node.anchors = anchors
        self.node.font_size = self.overlay.font_size

        self.x_size, self.y_size = self.node.get_width_height()

        # depending on the canvas position, we need to change the position of the anchor itself
        # to ensure the text aligns properly e.g. left when on the left, and right when on the right
        x = y = 0.0
        if anchors[0] == 'right':
            x = self.x_size
        elif anchors[0] == 'center':
            x = self.x_size / 2

        if anchors[1] == 'top':
            y = self.y_size

        self.node.pos = (x, y)

        super()._on_position_change()

    def reset(self):
        super().reset()
        self._on_text_change()
        self._on_color_change()



from unittest.mock import Mock

from napari.utils.task_status import Status, TaskStatusManager


def test_task_status():
    """test task status registration and update using the utils.task_status module."""
    task_status_manager = TaskStatusManager()

    # Check task status registration
    cancel_callback_mock = Mock()
    task_status_id = task_status_manager.register_task_status(
        'test-task-status',
        Status.BUSY,
        'Register task status busy',
        cancel_callback=cancel_callback_mock,
    )
    assert task_status_manager.is_busy()
    assert task_status_manager.get_status() == [
        'test-task-status: Register task status busy'
    ]

    # Check task status update
    task_status_manager.update_task_status(
        task_status_id,
        Status.COMPLETED,
        description='Register task status completed',
    )
    assert not task_status_manager.is_busy()
    assert task_status_manager.get_status() == []
    task_status_manager.update_task_status(
        task_status_id,
        Status.PENDING,
        description='Register task status pending',
    )
    assert task_status_manager.is_busy()
    assert task_status_manager.get_status() == [
        'test-task-status: Register task status pending'
    ]

    # Check cancel behavior
    task_status_manager.cancel_all()
    cancel_callback_mock.assert_called_once()
    assert task_status_manager.get_status() == []
    assert not task_status_manager.is_busy()



"""Tests of the Viewer class that interact directly with the Qt code"""

from __future__ import annotations

import os
import textwrap
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
from numpy import testing as npt
from qtpy.QtCore import QEvent, Qt, QUrl
from qtpy.QtGui import QGuiApplication, QKeyEvent
from qtpy.QtWidgets import QApplication

from napari._qt._tests.test_qt_viewer import qt_viewer
from napari._tests.utils import skip_local_popups, skip_on_win_ci
from napari.settings import get_settings
from napari.utils.theme import available_themes


@pytest.mark.usefixtures('builtins')
@pytest.mark.enable_console
@pytest.mark.filterwarnings('ignore::DeprecationWarning:jupyter_client')
@patch('qtpy.QtWidgets.QApplication.topLevelWidgets')
def test_drop_python_file(qapp_mock, make_napari_viewer, tmp_path):
    """Test dropping a python file on to the viewer."""
    viewer = make_napari_viewer()
    filename = tmp_path / 'image_to_drop.py'
    file_content = textwrap.dedent("""
    import numpy as np
    import napari

    data = np.zeros((10, 10))
    viewer = napari.Viewer()
    viewer.add_image(data, name='Dropped Image')

    napari.run()
    """)
    filename.write_text(file_content)

    # Simulate dropping the file
    mock_event = MagicMock()
    mock_event.mimeData.return_value.urls.return_value = [
        QUrl(filename.as_uri())
    ]
    viewer.window._qt_viewer.dropEvent(mock_event)

    # Check that the file was executed
    assert viewer.layers[0].name == 'Dropped Image'

    # Ensure that napari.run is stopped early and event loops are not started
    # by call napari.run() in the script.
    qapp_mock.assert_not_called()

    # Check that the console is updated with locals from the script
    console = viewer.window._qt_viewer.console
    assert 'data' in console.shell.user_ns
    assert np.array_equal(console.shell.user_ns['data'], np.zeros((10, 10)))


@pytest.mark.usefixtures('builtins')
@pytest.mark.enable_console
@pytest.mark.filterwarnings('ignore::DeprecationWarning:jupyter_client')
def test_drop_python_file_3d(make_napari_viewer, tmp_path):
    """Test that dropping a python file using a 3D image on the viewer works."""
    viewer = make_napari_viewer()
    filename = tmp_path / 'image_to_drop_3d.py'
    file_content = textwrap.dedent("""
    import numpy as np
    from napari import Viewer

    data = np.zeros((2, 10, 10))
    viewer = Viewer(ndisplay=3)
    viewer.add_image(data, name='Dropped Image')
    """)
    filename.write_text(file_content)

    # Simulate dropping the file
    mock_event = MagicMock()
    mock_event.mimeData.return_value.urls.return_value = [
        QUrl(filename.as_uri())
    ]
    viewer.window._qt_viewer.dropEvent(mock_event)

    # Check that the file was executed
    assert viewer.layers[0].name == 'Dropped Image'
    assert viewer.dims.ndim == 3

    # Check that the console is updated with locals from the script
    console = viewer.window._qt_viewer.console
    assert 'data' in console.shell.user_ns
    assert np.array_equal(console.shell.user_ns['data'], np.zeros((2, 10, 10)))


@pytest.mark.usefixtures('builtins')
@pytest.mark.enable_console
@pytest.mark.filterwarnings('ignore::DeprecationWarning:jupyter_client')
def test_drop_python_file_double_viewer(make_napari_viewer, tmp_path):
    """Test that dropping a python file on the viewer works."""
    viewer = make_napari_viewer()
    filename = tmp_path / 'test.py'
    file_content = textwrap.dedent("""
    import numpy as np
    from napari import Viewer

    data = np.zeros((10, 10))
    viewer1 = Viewer()
    viewer1.add_image(data, name='Dropped Image')
    viewer2 = Viewer(title="text")
    viewer2.add_image(data, name='Dropped Image 2')
    """)
    filename.write_text(file_content)

    # Simulate dropping the file
    mock_event = MagicMock()
    mock_event.mimeData.return_value.urls.return_value = [
        QUrl(filename.as_uri())
    ]
    viewer.window._qt_viewer.dropEvent(mock_event)

    # Check that the file was executed
    assert viewer.layers[0].name == 'Dropped Image'
    assert len(viewer._instances) == 2  # Two viewers should be created
    instances = list(viewer._instances)
    idx = 0 if instances[1] == viewer else 1
    assert instances[idx].title == 'text'  # Check the second viewer's name
    instances[idx].close()  # Close the second viewer

    # check that the console is updated with locals from the script
    console = viewer.window._qt_viewer.console
    assert 'data' in console.shell.user_ns
    assert np.array_equal(console.shell.user_ns['data'], np.zeros((10, 10)))
    assert 'viewer1' in console.shell.user_ns
    assert 'viewer2' in console.shell.user_ns


def test_qt_viewer(make_napari_viewer):
    """Test instantiating viewer."""
    viewer = make_napari_viewer()
    view = viewer.window._qt_viewer

    assert viewer.title == 'napari'
    assert view.viewer == viewer

    assert len(viewer.layers) == 0
    assert view.layers.model().rowCount() == 0

    assert viewer.dims.ndim == 2
    assert view.dims.nsliders == viewer.dims.ndim
    assert np.sum(view.dims._displayed_sliders) == 0


def test_qt_viewer_with_console(make_napari_viewer):
    """Test instantiating console from viewer."""
    viewer = make_napari_viewer()
    view = viewer.window._qt_viewer
    # Check console is created when requested
    assert view.console is not None
    assert view.dockConsole.widget() is view.console


def test_qt_viewer_toggle_console(make_napari_viewer):
    """Test instantiating console from viewer."""
    viewer = make_napari_viewer()
    view = viewer.window._qt_viewer
    # Check console has been created when it is supposed to be shown
    view.toggle_console_visibility(None)
    assert view._console is not None
    assert view.dockConsole.widget() is view.console


@skip_local_popups
@pytest.mark.skipif(os.environ.get('MIN_REQ', '0') == '1', reason='min req')
def test_qt_viewer_console_focus(qtbot, make_napari_viewer):
    """Test console has focus when instantiating from viewer."""
    viewer = make_napari_viewer(show=True)
    view = viewer.window._qt_viewer
    assert not view.console.hasFocus(), 'console has focus before being shown'

    view.toggle_console_visibility(None)

    def console_has_focus():
        assert view.console.hasFocus(), (
            'console does not have focus when shown'
        )

    qtbot.waitUntil(console_has_focus)


@skip_on_win_ci
def test_screenshot(make_napari_viewer, qapp, qtbot):
    "Test taking a screenshot"
    viewer = make_napari_viewer(show=True)
    rng = np.random.default_rng(0)
    # Add image
    data = rng.random((10, 15))
    viewer.add_image(data)

    # Add labels
    data = rng.integers(20, size=(10, 15))
    viewer.add_labels(data)

    # Add points
    data = 20 * rng.random((10, 2))
    viewer.add_points(data)

    # Add vectors
    data = 20 * rng.random((10, 2, 2))
    viewer.add_vectors(data)

    # Add shapes
    data = 20 * rng.random((10, 4, 2))
    viewer.add_shapes(data)
    # wait for window be fully rendered
    qtbot.wait_exposed(qt_viewer)

    # without these two lines, the shape of screenshot1 and screenshot2 differ by
    # 2 pixels. It looks like some event requires inactivity in event loop to
    # trigger resize to final shape.
    qtbot.wait(5)
    qapp.processEvents()

    screenshot2 = viewer.window.screenshot(flash=False, canvas_only=True)

    qapp.processEvents()

    # Take screenshot
    with pytest.warns(FutureWarning, match='qt_viewer'):
        screenshot1 = viewer.window.qt_viewer.screenshot(flash=False)

    npt.assert_array_equal(screenshot1, screenshot2)
    assert screenshot1.ndim == 3


@pytest.mark.slow
@skip_on_win_ci
def test_qt_viewer_clipboard_with_flash(make_napari_viewer, qtbot):
    viewer = make_napari_viewer()
    # make sure clipboard is empty
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()

    viewer.window._qt_viewer.clipboard(flash=True)

    viewer.window.clipboard(flash=False, canvas_only=True)

    clipboard_image = QGuiApplication.clipboard().image()
    assert not clipboard_image.isNull()

    # ensure the flash effect is applied
    assert (
        viewer.window._qt_viewer._welcome_widget.graphicsEffect() is not None
    )
    assert hasattr(
        viewer.window._qt_viewer._welcome_widget, '_flash_animation'
    )
    qtbot.wait(500)  # wait for the animation to finish
    assert viewer.window._qt_viewer._welcome_widget.graphicsEffect() is None
    assert not hasattr(
        viewer.window._qt_viewer._welcome_widget, '_flash_animation'
    )

    # clear clipboard and grab image from application view
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()

    # capture screenshot of the entire window
    viewer.window.clipboard(flash=True)
    clipboard_image = QGuiApplication.clipboard().image()
    assert not clipboard_image.isNull()

    # ensure the flash effect is applied
    assert viewer.window._qt_window.graphicsEffect() is not None
    assert hasattr(viewer.window._qt_window, '_flash_animation')
    qtbot.wait(500)  # wait for the animation to finish
    assert viewer.window._qt_window.graphicsEffect() is None
    assert not hasattr(viewer.window._qt_window, '_flash_animation')


@skip_on_win_ci
def test_qt_viewer_clipboard_without_flash(make_napari_viewer):
    viewer = make_napari_viewer()
    # make sure clipboard is empty
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()

    # capture screenshot
    viewer.window._qt_viewer.clipboard(flash=False)

    viewer.window.clipboard(flash=False, canvas_only=True)

    clipboard_image = QGuiApplication.clipboard().image()
    assert not clipboard_image.isNull()

    # ensure the flash effect is not applied
    assert viewer.window._qt_viewer._welcome_widget.graphicsEffect() is None
    assert not hasattr(
        viewer.window._qt_viewer._welcome_widget, '_flash_animation'
    )

    # clear clipboard and grab image from application view
    QGuiApplication.clipboard().clear()
    clipboard_image = QGuiApplication.clipboard().image()
    assert clipboard_image.isNull()

    # capture screenshot of the entire window
    viewer.window.clipboard(flash=False)
    clipboard_image = QGuiApplication.clipboard().image()
    assert not clipboard_image.isNull()

    # ensure the flash effect is not applied
    assert viewer.window._qt_window.graphicsEffect() is None
    assert not hasattr(viewer.window._qt_window, '_flash_animation')


@pytest.mark.parametrize('theme', available_themes())
def test_canvas_color(make_napari_viewer, theme):
    """Test instantiating viewer with different themes.

    See: https://github.com/napari/napari/issues/3278
    """
    # This test is to make sure the application starts with
    # with different themes
    get_settings().appearance.theme = theme
    viewer = make_napari_viewer()
    assert viewer.theme == theme


def test_shortcut_passing(make_napari_viewer):
    viewer = make_napari_viewer(ndisplay=3)
    layer = viewer.add_labels(
        np.zeros((2, 2, 2), dtype=np.uint8), scale=(1, 2, 4)
    )
    layer.mode = 'fill'

    qt_window = viewer.window._qt_window

    qt_window.keyPressEvent(
        QKeyEvent(
            QEvent.Type.KeyPress, Qt.Key.Key_1, Qt.KeyboardModifier.NoModifier
        )
    )
    qt_window.keyReleaseEvent(
        QKeyEvent(
            QEvent.Type.KeyPress, Qt.Key.Key_1, Qt.KeyboardModifier.NoModifier
        )
    )
    assert layer.mode == 'erase'


def test_points_2d_to_3d(make_napari_viewer):
    """See https://github.com/napari/napari/issues/6925"""
    # this requires a full viewer cause some issues are caused only by
    # qt processing events
    viewer = make_napari_viewer(ndisplay=2, show=True)
    viewer.add_points()
    QApplication.processEvents()
    viewer.dims.ndisplay = 3
    QApplication.processEvents()


def test_surface_mixed_dim(make_napari_viewer):
    """Test that adding a layer that changes the world ndim
    when ndisplay=3 before the mouse cursor has been updated
    doesn't raise an error.

    See PR: https://github.com/napari/napari/pull/3881
    """
    viewer = make_napari_viewer(ndisplay=3)

    verts = np.array([[0, 0, 0], [0, 20, 10], [10, 0, -10], [10, 10, -10]])
    faces = np.array([[0, 1, 2], [1, 2, 3]])
    values = np.linspace(0, 1, len(verts))
    data = (verts, faces, values)
    viewer.add_surface(data)

    timeseries_values = np.vstack([values, values])
    timeseries_data = (verts, faces, timeseries_values)
    viewer.add_surface(timeseries_data)



from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QComboBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari._qt.utils import qt_signals_blocked
from napari.layers import Vectors
from napari.layers.vectors._vectors_constants import VECTORSTYLE_TRANSLATIONS
from napari.utils.events.event_utils import connect_setattr
from napari.utils.translations import trans


class QtVectorStyleComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer vector style
    value attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Vectors
        An instance of a napari Vectors layer.

    Attributes
    ----------
    vector_style_combobox : qtpy.QtWidgets.QComboBox
        Dropdown widget to select vector_style for the vectors.
    vector_style_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for vector_style value chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Vectors) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.vector_style.connect(self._on_vector_style_change)

        # Setup widgets
        # dropdown to select the edge display vector_style
        vector_style_combobox = QComboBox(parent)
        for index, (data, text) in enumerate(VECTORSTYLE_TRANSLATIONS.items()):
            data = data.value
            vector_style_combobox.addItem(text, data)
            if data == self._layer.vector_style:
                vector_style_combobox.setCurrentIndex(index)

        self.vector_style_combobox = vector_style_combobox
        connect_setattr(
            self.vector_style_combobox.currentTextChanged,
            self._layer,
            'vector_style',
        )

        self.vector_style_combobox_label = QtWrappedLabel(
            trans._('vector style:')
        )

    def _on_vector_style_change(self) -> None:
        """Receive layer model vector style change event & update dropdown."""
        with qt_signals_blocked(self.vector_style_combobox):
            vector_style = self._layer.vector_style
            index = self.vector_style_combobox.findText(
                vector_style, Qt.MatchFlag.MatchFixedString
            )
            self.vector_style_combobox.setCurrentIndex(index)

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.vector_style_combobox_label, self.vector_style_combobox)]



from qtpy.QtWidgets import (
    QComboBox,
    QWidget,
)

from napari._qt.layer_controls.widgets.qt_widget_controls_base import (
    QtWidgetControlsBase,
    QtWrappedLabel,
)
from napari.layers import Labels
from napari.layers.labels._labels_constants import (
    LABEL_COLOR_MODE_TRANSLATIONS,
    LabelColorMode,
)
from napari.utils import CyclicLabelColormap
from napari.utils.translations import trans


class QtColorModeComboBoxControl(QtWidgetControlsBase):
    """
    Class that wraps the connection of events/signals between the layer color
    mode attribute and Qt widgets.

    Parameters
    ----------
    parent: qtpy.QtWidgets.QWidget
        An instance of QWidget that will be used as widgets parent
    layer : napari.layers.Labels
        An instance of a napari Labels layer.

    Attributes
    ----------
    color_mode_combobox : qtpy.QtWidgets.QComboBox
        ComboBox controlling current color mode of the layer.
    color_mode_combobox_label : napari._qt.layer_controls.widgets.qt_widget_controls_base.QtWrappedLabel
        Label for the color mode chooser widget.
    """

    def __init__(self, parent: QWidget, layer: Labels) -> None:
        super().__init__(parent, layer)
        # Setup layer
        self._layer.events.colormap.connect(self._on_colormap_change)

        # Setup widgets
        color_mode_comboBox = QComboBox()
        for data, text in LABEL_COLOR_MODE_TRANSLATIONS.items():
            data = data.value
            color_mode_comboBox.addItem(text, data)

        self.color_mode_combobox = color_mode_comboBox
        self._on_colormap_change()
        color_mode_comboBox.activated.connect(self.change_color_mode)

        self.color_mode_combobox_label = QtWrappedLabel(trans._('color mode:'))

    def change_color_mode(self) -> None:
        """Change color mode of label layer"""
        if self.color_mode_combobox.currentData() == LabelColorMode.AUTO.value:
            self._layer.colormap = self._layer._original_random_colormap
        else:
            self._layer.colormap = self._layer._direct_colormap

    def _on_colormap_change(self) -> None:
        enable_combobox = not self._layer._is_default_colors(
            self._layer._direct_colormap.color_dict
        )
        self.color_mode_combobox.setEnabled(enable_combobox)
        if not enable_combobox:
            self.color_mode_combobox.setToolTip(
                'Layer needs a user-set DirectLabelColormap to enable direct '
                'mode.'
            )
        if isinstance(self._layer.colormap, CyclicLabelColormap):
            self.color_mode_combobox.setCurrentIndex(
                self.color_mode_combobox.findData(LabelColorMode.AUTO.value)
            )
        else:
            self.color_mode_combobox.setCurrentIndex(
                self.color_mode_combobox.findData(LabelColorMode.DIRECT.value)
            )

    def get_widget_controls(self) -> list[tuple[QtWrappedLabel, QWidget]]:
        return [(self.color_mode_combobox_label, self.color_mode_combobox)]



from typing import TYPE_CHECKING

from napari._qt.layer_controls.qt_image_controls_base import (
    QtBaseImageControls,
)
from napari._qt.layer_controls.widgets import QtProjectionModeControl
from napari._qt.layer_controls.widgets._image import (
    QtDepictionControl,
    QtImageRenderControl,
    QtInterpolationComboBoxControl,
)

if TYPE_CHECKING:
    import napari.layers


class QtImageControls(QtBaseImageControls):
    """Qt view and controls for the napari Image layer.

    Parameters
    ----------
    layer : napari.layers.Image
        An instance of a napari Image layer.

    Attributes
    ----------
    _depiction_control : napari._qt.layer_controls.widgets._image.QtDepictionControl
        Widget that wraps widgets related with the layer depiction and plane attributes.
    _interpolation_control : napari._qt.layer_controls.widgets._image.QtInterpolationComboBoxControl
        Widget that wraps dropdown menu to select the interpolation mode for image display.
    _projection_mode_control : napari._qt.layer_controls.widgets.QtProjectionModeControl
        Widget that wraps dropdown menu to select the projection mode for the layer.
    _render_control : napari._qt.layer_controls.widgets._image.QtImageRenderControl
        Widget that wraps widgets related with the method used to render the layer.
    """

    layer: 'napari.layers.Image'
    PAN_ZOOM_ACTION_NAME = 'activate_image_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_image_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)
        # Setup widgets controls
        self._projection_mode_control = QtProjectionModeControl(self, layer)
        self._add_widget_controls(self._projection_mode_control)
        self._interpolation_control = QtInterpolationComboBoxControl(
            self, layer
        )
        self._add_widget_controls(self._interpolation_control)
        self._depiction_control = QtDepictionControl(self, layer)
        self._add_widget_controls(self._depiction_control)
        self._render_control = QtImageRenderControl(self, layer)
        self._add_widget_controls(self._render_control)

        self._on_ndisplay_changed()

    def _on_ndisplay_changed(self):
        """Update widget visibility based on 2D and 3D visualization modes."""
        self._interpolation_control._update_interpolation_combo(self.ndisplay)
        self._depiction_control._update_plane_parameter_visibility()
        if self.ndisplay == 2:
            self._render_control._on_display_change_hide()
            self._depiction_control._on_display_change_hide()
        else:
            self._render_control._on_display_change_show()
            self._depiction_control._on_display_change_show()
        super()._on_ndisplay_changed()
